import unittest
import os
import io

from pandas.util.testing import assert_frame_equal

from oplus.configuration import CONF
from oplus.epw import parse_epw, EPW

weather_dir = os.path.join(CONF.eplus_base_dir_path, "WeatherData")
resources_dir = os.path.join(os.path.realpath(os.path.dirname(__file__)), "resources", "epw")


class EPWTestCase(unittest.TestCase):
    def assert_epw_equal(self, expected_content, given_content):
        expected_content_l2 = [[cell.strip() for cell in row.split(",")] for row in expected_content.split("\n")]
        given_content_l2 = [[cell.strip() for cell in row.split(",")] for row in given_content.split("\n")]

        for r, expected_row in enumerate(expected_content_l2):
            for c, expected_cell in enumerate(expected_row):
                try:
                    self.assertEqual(float(expected_cell), float(given_content_l2[r][c]))
                except ValueError:
                    self.assertEqual(expected_cell, given_content_l2[r][c], "Cells differ -> row: %i, column: %i" %
                                     (r, c))


class EPlusWeatherData(EPWTestCase):
    """
    Tested under EPlus 8.1.0 on Windows (Geoffroy).
    """

    def test_epw_to_df_to_epw_integrity(self):
        """
        tests if epw to df to epw works
        """
        epw_files_l = []
        for file_name in os.listdir(weather_dir):
            if os.path.splitext(file_name)[1] == ".epw":
                epw_files_l.append(file_name)

        for file_name in epw_files_l:
            file_path = os.path.join(weather_dir, file_name)
            with open(file_path) as f:
                expected_content = f.read()

            epw = EPW(file_path)
            f = io.StringIO()
            epw.save_as(f, add_copyright=False)

            new_content = f.getvalue()

            self.assert_epw_equal(expected_content, new_content)

    def test_df_get_and_set_integrity(self):
        epw = EPW(os.path.join(weather_dir, "USA_CA_San.Francisco.Intl.AP.724940_TMY3.epw"))
        df = epw.df()
        epw.set_df(df)
        assert_frame_equal(df, epw.df())



"""
Useful functions
--------------------


"""

import datetime as dt
import copy
import logging
import subprocess
import collections
import os
import io
import sys
import threading
import contextlib


from oplus import __version__, CONF


logger = logging.getLogger(__name__)


class UtilError(Exception):
    pass


class CachingNotAllowedError(Exception):
    pass


def get_copyright_comment(multi_lines=True):
    if multi_lines:
#         return """----------------------------------------------------------------------------------------
#  |----| |----| |      |    |  |----|      Generated by OPlus version %s.
#  |    | |    | |      |    |  |           Copyright (c) %s, Openergy development team
#  |    | |----| |      |    |  |----|      http://www.openergy.fr
#  |    | |      |      |    |       |      https://github.com/Openergy/oplus
#  |----| |      |----| |----|  |----|
# ----------------------------------------------------------------------------------------
#
# """
        return """----------------------------------------------------------------------------------------
Generated by OPlus version %s.
Copyright (c) %s, Openergy development team
http://www.openergy.fr
https://github.com/Openergy/oplus
----------------------------------------------------------------------------------------\n""" % (
            __version__,
            dt.datetime.now().year
        )
    return "-- OPlus version %s, copyright (c) %s, Openergy development team --" % (__version__, dt.datetime.now().year)


class EPlusDt:
    MONTH = 0
    DAY = 1
    HOUR = 2
    MINUTE = 3

    @classmethod
    def from_datetime(cls, datetime):
        if datetime.minute == 0:
            datetime -= dt.timedelta(hours=1)
            minute = 60
        else:
            minute = datetime.minute

        return cls(datetime.month, datetime.day, datetime.hour + 1, minute)

    @classmethod
    def to_datetime(cls, year, month, day, eplus_hour, eplus_minute):
        """
        Arguments
        ---------
        month
        day
        hour (1 -> 24) => one hour shift
        minute (1 -> 60) => no shift, but 0 is 60 one hour before. We tolerate minute=0 even though eplus does not
            seem to use it.
        """
        hour_cor = 0
        if eplus_minute == 60:
            minute = 0
            hour_cor = 1
        else:
            minute = eplus_minute

        try:
            my_dt = (dt.datetime(year, month, day, eplus_hour-1, minute) +
                     dt.timedelta(hours=hour_cor)).replace(year=year)  # we replace year for one case: y-12-31-24-60
            my_dt.replace(year=year)  # we replace in case timedelta operation impacted year
        except ValueError as e:
            if (month, day) == (2, 29):
                raise UtilError("%s (probable leap year problem: year=%s, month=%s, day=%s)" % (e, year, month, day))
            raise e

        return my_dt

    def __init__(self, month, day, out_hour, out_minute):
        """
        Arguments
        ---------
        month
        day
        hour (1 -> 24) => one hour shift
        minute (1 -> 60) => no shift, but 0 is 60 one hour before. We tolerate minute=0 even though eplus does not
        # todo: unify hour/minute convention, if possible
            seem to use it.
        """
        self._standard_dt = self.to_datetime(2000, month, day, out_hour, out_minute)  # 2000 is a leap year

        # create and store value
        _datetime = copy.copy(self._standard_dt)
        if self._standard_dt.minute == 0:
            _datetime -= dt.timedelta(hours=1)
            minute = 60
        else:
            minute = _datetime.minute

        self._value = _datetime.month, _datetime.day, _datetime.hour + 1, minute

    def __lt__(self, other):
        return self.standard_dt < other.standard_dt

    def __le__(self, other):
        return self.standard_dt <= other.standard_dt

    def __eq__(self, other):
        return self.standard_dt == other.standard_dt

    def __ne__(self, other):
        return self.standard_dt != other.standard_dt

    def __gt__(self, other):
        return self.standard_dt > other.standard_dt

    def __ge__(self, other):
        return self.standard_dt >= other.standard_dt

    def __repr__(self):
        return "<eplus_dt: month=%s, day=%s, hour=%s, minute=%s>" % self._value

    def datetime(self, year):
        return self.to_datetime(year, self._value[self.MONTH], self._value[self.DAY],
                                self._value[self.HOUR], self._value[self.MINUTE])

    @property
    def month(self):
        return self._value[self.MONTH]

    @property
    def day(self):
        return self._value[self.DAY]

    @property
    def hour(self):
        return self._value[self.HOUR]

    @property
    def minute(self):
        return self._value[self.MINUTE]

    @property
    def standard_dt(self):
        return self._standard_dt


def get_start_dt(start):
    """
    Transforms start in start_dt.

    Arguments
    ---------
    start: year num, or date, or datetime
    """
    if isinstance(start, dt.datetime):  # must first test datetime because date is datetime...
        start_dt = start
    elif isinstance(start, dt.date):
        start_dt = dt.datetime(start.year, start.month, start.day)
    elif isinstance(start, int):
        start_dt = dt.datetime(start, 1, 1)
    else:
        raise UtilError("Unknown start type: '%s'." % type(start))
    return start_dt


def _redirect_stream(src, dst, stop_event, freq):
    while not stop_event.is_set():  # read all filled lines
        content = src.readline()
        if content == "":  # empty: break
            break
        dst.write(content)
        if hasattr(dst, "flush"):
            dst.flush()


@contextlib.contextmanager
def redirect_stream(src, dst, freq=0.1):
    stop_event = threading.Event()
    t = threading.Thread(target=_redirect_stream, args=(src, dst, stop_event, freq))
    t.daemon = True
    t.start()
    try:
        yield
    finally:
        stop_event.set()
        t.join()


class LoggerStreamWriter:
    def __init__(self, logger_name, level):
        self._logger = logging.getLogger(logger_name)
        self._level = level

    def write(self, message):
        message = message.strip()
        if message != "":
            self._logger.log(self._level, message)


def run_subprocess(command, cwd=None, stdout=None, stderr=None, shell=False, beat_freq=None):
    """
    Parameters
    ----------
    command: command
    cwd: current working directory
    stdout: output info stream (must have 'write' method)
    stderr: output error stream (must have 'write' method)
    shell: see subprocess.Popen
    beat_freq: if not none, stdout will be used at least every beat_freq (in seconds)
    """
    # prepare variables
    stdout = sys.stdout if stdout is None else stdout
    stderr = sys.stderr if stderr is None else stderr

    # run subprocess
    with subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=cwd,
        shell=shell,
        universal_newlines=True
    ) as sub_p:
        # link output streams
        with redirect_stream(sub_p.stdout, stdout), redirect_stream(sub_p.stderr, stderr):
            while True:
                try:
                    sub_p.wait(timeout=beat_freq)
                    break
                except subprocess.TimeoutExpired:
                    stdout.write("subprocess is still running\n")
                    if hasattr(sys.stdout, "flush"):
                        sys.stdout.flush()
        return sub_p.returncode


class CacheKey:
    """
    emulated a dict that can store hashable types
    """
    def __init__(self, method, *args, **kwargs):
        self._value = tuple([method] + list(args) + [(k, v) for k, v in sorted(kwargs.items())])

    def __hash__(self):
        return self._value.__hash__()

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __str__(self):
        return "<CacheKey: %s>" % str(self._value)


def check_cache_is_off(method):
    def wrapper(self, *args, **kwargs):
        assert isinstance(self, Cached), "decorator was applied to a non-cached class (%s)" % method
        if self.is_cached:
            raise CachingNotAllowedError("Must turn off cache to perform action.")
        return method(self, *args, **kwargs)
    return wrapper


def cached(method):
    def wrapper(self, *args, **kwargs):
        assert isinstance(self, Cached), "decorator was applied to a non-cached class (%s)" % method
        if not self.is_cached:
            return method(self, *args, **kwargs)
        key = CacheKey(method, *args, **kwargs)
        if key not in self.cache:
            self.cache[key] = dict(value=method(self, *args, **kwargs), hits=0)
        else:
            self.cache[key]["hits"] += 1
        return self.cache[key]["value"]
    return wrapper


class Cached:
    cache = None  # dict(key: dict(value=v, hits=0))  (hits for testing)

    def activate_cache(self):
        if self.cache is None:
            self.cache = {}

    def deactivate_cache(self):
        self.cache = None

    def clear_cache(self):
        if self.cache is not None:
            self.cache = {}

    @property
    def is_cached(self):
        return self.cache is not None


class Enum(collections.UserDict):
    def __getattr__(self, item):
        return self[item]

    def __setattr__(self, key, value):
        # we bypass 'data' (used by parent class) and all private keys (in case used by parent class)
        if (key == "data") or (key[0] == "_"):
            super().__setattr__(key, value)
        self[key] = value


def get_string_buffer(path_or_content, expected_extension, encoding):
    """
    path_or_content: path or content_str or content_bts or string_io or bytes_io

    Returns
    -------
    string_buffer, path

    path will be None if input was not a path
    """
    buffer, path = None, None

    # path or content string

    if isinstance(path_or_content, str):
        if path_or_content[-len(expected_extension)-1:] == ".%s" % expected_extension:
            assert os.path.isfile(path_or_content), "No file at given path: '%s'." % path_or_content
            buffer, path = open(path_or_content, encoding=encoding), path_or_content
        else:
            buffer = io.StringIO(path_or_content, )

    # text io
    elif isinstance(path_or_content, io.TextIOBase):
        buffer = path_or_content

    # bytes
    elif isinstance(path_or_content, bytes):
        buffer = io.StringIO(path_or_content.decode(encoding=encoding))
    elif isinstance(path_or_content, io.BufferedIOBase):
        buffer = io.StringIO(path_or_content.read().decode(encoding=encoding))
    else:
        raise UtilError("path_or_content type could not be identified")

    return buffer, path


class IDFStyle:
    head_key = None
    chapter_key = None
    object_key = None
    tail_object_key = None
    tail_type = None

    def get_chapter_title(self, content):
        # default: oplus
        length = 100
        s = self.get_chapter_comment("#" * length)
        first_length = length // 2 - len(content) // 2 - 1
        middle_string = "#" * first_length
        middle_string += " " + content.upper() + " "
        middle_string += "#" * (length - first_length - len(content) - 2)
        s += self.get_chapter_comment(middle_string)
        s += self.get_chapter_comment("#" * length)
        return s

    def get_object_comment(self, comment, line_jump=True):
        prefix = "!" + self.object_key
        if self.object_key != " ":
            prefix += " "
        s = prefix + comment
        if line_jump:
            s += "\n"
        return s

    def get_tail_object_comment(self, comment, line_jump=True):
        prefix = "!" + self.tail_object_key
        if self.tail_object_key != " ":
            prefix += " "
        s = prefix + comment
        if line_jump:
            s += "\n"
        return s

    def get_head_comment(self, comment, line_jump=True):
        prefix = "!" + self.head_key
        if self.head_key != " ":
            prefix += " "
        s = prefix + comment
        if line_jump:
            s += "\n"
        return s

    def get_chapter_comment(self, comment, line_jump=True):
        prefix = "!" + self.chapter_key
        if self.chapter_key != " ":
            prefix += " "
        s = prefix + comment
        if line_jump:
            s += "\n"
        return s


class OplusIDFStyle(IDFStyle):
    head_key = " "
    chapter_key = "#"
    object_key = "-"
    tail_object_key = "-"
    tail_type = "after"


class DefaultWriteIDFStyle(IDFStyle):
    head_key = " "
    chapter_key = " "
    object_key = " "
    tail_object_key = " "
    tail_type = "after"


class DesignBuilderIDFStyle(IDFStyle):
    head_key = " "
    chapter_key = "#"
    object_key = "-"
    tail_object_key = " "
    tail_type = "before"


class ASHRAEIDFStyle(IDFStyle):
    head_key = " "
    chapter_key = " -"
    object_key = "  -"
    tail_object_key = " "
    tail_type = "before"

    def get_chapter_title(self, content):
        s = " "*2 + "="*11 + " "*2
        s += "ALL OBJECTS IN CLASS: " + content.upper()
        s += " " + "="*11

        return self.get_chapter_comment(s)


style_library = {
    None: None,
    "oplus": OplusIDFStyle(),
    "ASHRAE": ASHRAEIDFStyle(),
    "default write": DefaultWriteIDFStyle(),
    "design builder": DesignBuilderIDFStyle(),
}



"""
Summary table
---------------


"""

import os
import re
import io

import pandas as pd

from oplus.configuration import CONF


class SummaryTableError(Exception):
    pass


class SummaryTable:
    def __init__(self, path, encoding=None):
        assert os.path.isfile(path), "No file at given path: '%s'." % path
        self.path = path
        self.encoding = CONF.encoding if encoding is None else encoding

        self.sep = None
        self.report_tables_ref = {}

        self._parse()

    def _parse(self):
        with open(self.path, encoding=self.encoding) as f:
            start_parse = False
            search_end = False
            for var in enumerate(f):
                line_nb = var[0]
                line_s = var[1]

                if 'Tabular Output Report in Format' in line_s:
                    self.sep = line_s.split(':')[1][1]

                if 'REPORT:' in line_s:
                    start_parse = True
                    if search_end:
                        self.report_tables_ref['{r}_{f}'.format(r=report, f=for_)][table_name]['lineend'] = line_nb-1
                        search_end = False
                    report = re.findall('([^%s]+)' % self.sep, line_s)[1].replace('\n', '')
                elif not start_parse:
                    continue
                elif 'FOR:' in line_s:
                    for_ = re.findall('([^%s]+)' % self.sep, line_s)[1].replace('\n', '')
                    self.report_tables_ref['{r}_{f}'.format(r=report, f=for_)] = {}
                    self.report_tables_ref['{r}_{f}'.format(r=report, f=for_)]['TableListName'] = []
                    continue

                elif not any(
                        [v in line_s for v in
                         ['Values gathered over',
                          'WARNING:',
                          'Note',
                          '----',
                          'Values in table are in hours.'
                          ]]) and line_s[0] != '%s' % self.sep and line_s[0:2] != '\n':
                    if search_end:
                        self.report_tables_ref['{r}_{f}'.format(r=report, f=for_)][table_name]['lineend'] = line_nb-1
                        search_end = False
                    table_name = line_s.split('%s' % self.sep)[0].replace('\n', '')
                    if table_name not in self.report_tables_ref['{r}_{f}'.format(r=report, f=for_)]['TableListName']:
                        self.report_tables_ref['{r}_{f}'.format(r=report, f=for_)]['TableListName'].append(table_name)
                    self.report_tables_ref['{r}_{f}'.format(r=report, f=for_)][table_name] = {}

                elif (
                    (line_s.split('%s' % self.sep)[0] == '') and
                    (line_s.split('%s' % self.sep)[1] == '') and
                    (line_s.split('%s' % self.sep)[2] != '') and
                    ('linestart' not in self.report_tables_ref['{r}_{f}'.format(
                        r=report, f=for_)][table_name].keys())
                ):
                    self.report_tables_ref['{r}_{f}'.format(r=report, f=for_)][table_name]['linestart'] = line_nb
                    search_end = True

        # TODO: Manage key error correctly
        delete_t = []
        for report_key in self.report_tables_ref.keys():
            for table_key in self.report_tables_ref[report_key].keys():
                if table_key == 'TableListName':
                    continue
                if 'linestart' not in self.report_tables_ref[report_key][table_key].keys():
                    delete_t.append((report_key, table_key))
        # remove table ref
        for (r_key, t_key) in delete_t:
            del self.report_tables_ref[r_key][t_key]
            self.report_tables_ref[r_key]['TableListName'].remove(t_key)

    def get_report_keys(self):
        return list(self.report_tables_ref.keys())

    def get_table_report_list(self, report_key):
        return self.report_tables_ref[report_key]['TableListName']

    def get_table_df(self, report_key, table_report):
        content_bytes = open(self.path, "rb").read()
        content = content_bytes.decode(self.encoding).encode('ascii', 'ignore')
        f = io.BytesIO(content)

        begin_line = self.report_tables_ref[report_key][table_report]['linestart']
        end_line = self.report_tables_ref[report_key][table_report]['lineend']

        df = pd.read_csv(
            f,
            #sep=self.sep,
            sep=None,
            skiprows=begin_line,
            nrows=end_line-begin_line-3,
            index_col=1
            )
        df = df.dropna(axis='columns', how='all')
        df = df.dropna(axis='rows', how='all')

        # delete index name
        df.index.name = None

        return df

if __name__ == '__main__':
    import oplus as op
    rsc_path = os.path.join(os.getcwd().split('oplus')[0], 'antoine-work', 'csv')

    summary = SummaryTable(os.path.join(rsc_path, 'eplustbl.csv'))
    # for k in summary.get_report_keys():
    #     if 'Sizing' in k:
    #         print(k)

    # for tablename in summary.get_table_report_list('Component Sizing Summary_Entire Facility'):
    #     print(tablename)

    # for tablename in summary.get_table_report_list('Component Sizing Summary_Entire Facility'):
    #     print(tablename)
    #     df = summary.get_table_df('Component Sizing Summary_Entire Facility', tablename)

    #
    # with open(os.path.join(rsc_path, 'eplustbl.csv'), encoding=CONF.encoding) as f:
    #     for var in enumerate(f):
    #         if var[0] > 8735 and var[0] <= 9091:
    #             print(var[1])

    df = summary.get_table_df('Component Sizing Summary_Entire Facility', 'AirTerminal:SingleDuct:Uncontrolled')
    print(df)


import os
import unittest
import datetime as dt
import tempfile

from oplus.idf import IDF
from oplus.configuration import CONF
from oplus.simulation import simulate


class ZoneWithShadingSimple1(unittest.TestCase):
    """
    Tested under EPlus 8.1.0 on Windows (Geoffroy).
    """
    idf_path = os.path.join(CONF.eplus_base_dir_path, "ExampleFiles", "4ZoneWithShading_Simple_1.idf")
    epw_path = os.path.join(CONF.eplus_base_dir_path, "WeatherData", "USA_CO_Golden-NREL.724666_TMY3.epw")

    def add_output(self, idf, time_step):
        idf.add_object("Output:Variable,*,Site Outdoor Air Drybulb Temperature,%s;" % time_step)

    def test_all_time_steps(self):
        """
        Detailed doesn't work, but not tested because not a priority. # todo: understand detailed
        """
        time_steps_l = ["TimeStep", "Hourly", "Daily", "Monthly", "RunPeriod"]
        for time_step in time_steps_l:
            with tempfile.TemporaryDirectory() as temp_dir_path:
                idf = IDF(self.idf_path)
                self.add_output(idf, time_step)
                s = simulate(idf, self.epw_path, temp_dir_path)
                self.assertIsNotNone(s.eso.df(time_step=time_step), "Problem for: '%s'." % time_step)

    def test_start_dt(self):
        with tempfile.TemporaryDirectory() as temp_dir_path:
            idf = IDF(self.idf_path)
            self.add_output(idf, "Hourly")
            s = simulate(idf, self.epw_path, temp_dir_path, simulation_control="RunPeriods")
            start_dt = dt.datetime(2000, 1, 1)
            self.assertEqual(s.eso.df(time_step="Hourly", start=start_dt).index[0], dt.datetime(2000, 1, 1, 1))


"""
EPW
---

Objects describing all the complexity of an EPW file
"""
import os
import io
import re

import pandas as pd
from pandas.util.testing import assert_index_equal

from oplus.configuration import CONF
from oplus.util import EPlusDt, get_start_dt, get_copyright_comment, get_string_buffer


class EPWError(Exception):
    pass


EPW_COLUMNS = (
    "year",
    "month",
    "day",
    "hour",
    "minute",
    "datasource",
    "drybulb",
    "dewpoint",
    "relhum",
    "atmos_pressure",
    "exthorrad",
    "extdirrad",
    "horirsky",
    "glohorrad",
    "dirnorrad",
    "difhorrad",
    "glohorillum",
    "dirnorillum",
    "difhorillum",
    "zenlum",
    "winddir",
    "windspd",
    "totskycvr",
    "opaqskycvr",
    "visibility",
    "ceiling_hgt",
    "presweathobs",
    "presweathcodes",
    "precip_wtr",
    "aerosol_opt_depth",
    "snowdepth",
    "days_last_snow",
    "albedo",
    "liq_precip_depth",
    "liq_precip_rate"
)

WEEK_DAYS = (
    "Monday",
    "Tuesday",
    "Wednesday",
    "Thursday",
    "Friday",
    "Saturday",
    "Sunday"
)


class EPWHeader:
    """
    Representing the beginning of EPW files. It gathers information about the building's environment.
    """
    fields_d = {
        'LOCATION': 0,
        'DESIGN_CONDITIONS': 1,
        'TYPICAL_EXTREME_PERIODS': 2,
        'GROUND_TEMPERATURES': 3,
        'HOLIDAYS_DAYLIGHT_SAVINGS': 4,
        'COMMENTS_1': 5,
        'COMMENTS_2': 6,
        'DATA_PERIODS': 7,
    }

    @staticmethod
    def copyright_comment():
        return get_copyright_comment(multi_lines=False)

    def __init__(self, header_s):
        self._l2 = []
        for line_s in header_s.strip().split("\n"):
            self._l2.append([cell.strip() for cell in line_s.split(",")])
        # check
        if len(self._l2[self.fields_d['DATA_PERIODS']]) != 7:
            raise EPWError("'DATA PERIODS' row must have 7 cells.")
        if self._l2[self.fields_d['DATA_PERIODS']][0] != "DATA PERIODS":
            raise EPWError("Last line of header must be 'DATA PERIODS'.")
        if self._l2[self.fields_d['DATA_PERIODS']][1] != "1":
            raise NotImplementedError("Can only manage epws with one data period.")
        if self._l2[self.fields_d['DATA_PERIODS']][2] != "1":
            raise NotImplementedError("Can only manage hourly epws.")

    def to_str(self, add_copyright=True):
        l2 = []
        if add_copyright:
            for i, row in enumerate(self._l2):
                if add_copyright and (i == self.fields_d['COMMENTS_1']):
                    copyright_comment = self.copyright_comment()
                    row = row.copy()
                    if copyright_comment not in row[1]:
                        row[1] = "%s; %s" % (copyright_comment, row[1])
                l2.append(row)
        else:
            l2 = self._l2
        return "\n".join([",".join(row) for row in l2])

    @property
    def start_day_of_week(self):
        return self._l2[self.fields_d['DATA_PERIODS']][4].strip()

    @start_day_of_week.setter
    def start_day_of_week(self, value):
        # TODO: check given value is ok
        self._l2[self.fields_d['DATA_PERIODS']][4] = value

    @property
    def start(self):  # TODO: rename to start day
        """
        Access to the starting day of EPW files

        Returns
        -------
        Starting day
        """
        return self._l2[self.fields_d['DATA_PERIODS']][5].split(",").map(int)

    @start.setter
    def start(self, value):
        self._l2[self.fields_d['DATA_PERIODS']][5] = "%s/%s" % value

    @property
    def end(self):  # TODO: rename to end_day
        """

        Returns
        -------
        The end date of the EPW file (month, day)
        """
        return self._l2[self.fields_d['DATA_PERIODS']][6].split(",").map(int)

    @end.setter
    def end(self, value):
        """
        Set the new end date of the EPW file

        Parameters
        ----------
        value: (int, int)
            representing the end date (month, day)

        Returns
        -------

        """
        self._l2[self.fields_d['DATA_PERIODS']][6] = "%s/%s" % value

    @property
    def freq(self):
        return {"1": "H"}[self._l2[self.fields_d['DATA_PERIODS']][2]]

    def get_field(self, field):
        """

        Parameters
        ----------
        field: this field should be a key of the fields_d dictionary

        Returns
        -------
        A list containing the parameters' value of the defined field in the header
        """
        return self._l2[self.fields_d[field]][1:]

    def set_field(self, field, data_l=None, parameter_index=None, value=None):
        """
        This function sets the defined field of the header. Two choices :
         - put in parameter the entire new list
         - put the index and value of the parameter

        Parameters
        ----------
        field: this field should be a key of the fields_d dictionary
        data_l: list corresponding to the field. It has to fit the requirements
        parameter_index: first parameter is 1
        value: value of the parameter

        Returns
        -------

        """
        if field in self.fields_d.keys():
            if data_l:
                if data_l[0] != field.replace('_', ' '):
                    data_l.insert(0, field.replace('_', ' '))
                self._l2[self.fields_d[field]] = data_l
            elif parameter_index:
                if parameter_index == 0:
                    raise ValueError('1 is the first index for parameters')
                if value:
                    try:
                        self._l2[self.fields_d[field]][parameter_index] = value
                    except IndexError:
                        print('')

                else:
                    raise ValueError('If a parameter index is specified a value of the parameter is needed')
            else:
                raise ValueError('You have to specify a list of data or parameter index to set')
        else:
            raise KeyError('The mentioned field is not in header fields')

    def clear_field(self, field):
        """
        Clear the list of the corresponding field

        Parameters
        ----------
        field: this field should be a key of the fields_d dictionary

        Returns
        -------

        """
        if field in self.fields_d.keys():
            self._l2[self.fields_d[field]] = [field.replace('_', ' ')]
        else:
            raise KeyError('The mentioned field is not in header fields')



class EPW:
    """
    EPW
    ---

    Manage EPW files through EPW objects
    """
    epw_header_cls = EPWHeader

    @classmethod
    def get_epw_or_path(cls, epw_or_path, encoding=None):
        if isinstance(epw_or_path, str):
            return cls(epw_or_path, encoding=encoding)
        elif isinstance(epw_or_path, cls):
            return epw_or_path
        raise EPWError("'epw_or_path' must be an EPW or path.  Given object: '%s', type: '%s'." %
                       (epw_or_path, type(epw_or_path)))

    def __init__(self, path_or_buffer, encoding=None, start=None):
        """
        Initialize EPW objects

        Parameters
        ----------
        path_or_buffer: :obj: `os.path`
            path leading to the EPW file. It must end by .epw
        encoding:
        start:
        """
        self._encoding = CONF.encoding if encoding is None else encoding
        buffer, path = get_string_buffer(path_or_buffer, "epw", self._encoding)

        try:
            self._df, self._header = parse_epw(buffer)
        except Exception as e:
            raise EPWError("Error while parsing epw. First check that given file exists"
                           "(if not, given path will have been considered as an idf content.\n%s" % e)

        self._start_dt = None if start is None else get_start_dt(start)

    @property
    def header(self):
        return self._header

    @property
    def freq(self):
        return self._header.freq

    def to_str(self, add_copyright=True):
        # header
        _content = self._header.to_str(add_copyright=add_copyright)

        # data
        _f = io.StringIO()
        self._df.reset_index().to_csv(_f, header=False, index=False)
        _content += "\n" + _f.getvalue()

        return _content

    def save_as(self, file_or_path, add_copyright=True):
        """
        It enables to save an EPW file from EPW objects

        Parameters
        ----------
        file_or_path: os.path
            path of the EPW file. If it doesn't exist it would be created.
        add_copyright: bool
            option to add a copyright field. default value is True

        Returns
        -------

        """
        is_path = isinstance(file_or_path, str)

        # write to f
        f = (open(file_or_path, "w", encoding=self._encoding)
             if is_path else file_or_path)
        f.write(self.to_str(add_copyright=add_copyright))

        if is_path:
            f.close()

    def set_start(self, start):
        """
        Set the start day of the year for the EPW object. The end day would be defined automatically.
        An epw has to be one year long.

        Parameters
        ----------
        start: defines the start date. You can use a year date (int), a date (datetime.date),
         or a datetime (datetime.datetime)

        Returns
        -------

        """
        self._start_dt = get_start_dt(start)

    def df(self, start=None, datetime_index=None):
        """
        This function is a way to access weather data so that you can work on it more easily

        Parameters
        ----------
        start : optional
            option to define the starting date of the dataframe. If None it would be the start day attribute of
        datetime_index: str
            hello guys

        Returns
        -------
        :obj: `pandas.DataFrame()`
            The dataframe representing the weather data of the EPW file
        """
        # Order: will impose a freq, should not be used with reference climate (where years may differ by month)
        # start_dt and datetime index
        if start is None:
            start_dt = self._start_dt
        else:
            start_dt = get_start_dt(start)

        if datetime_index is None:
            datetime_index = start_dt is not None
        if (datetime_index is True) and (start_dt is None):
            # we get first value as start_dt
            i0 = self._df.index[0]
            start_dt = EPlusDt(i0[1], i0[2], i0[3], i0[4]).datetime(i0[0])

        # data frame
        df = self._df.copy()

        if not datetime_index:  # return if multi-index
            return df

        # manage datetime df
        start_standard_dt = EPlusDt.from_datetime(start_dt).standard_dt

        def index_to_dt(i):
            eplusdt = EPlusDt(i[1], i[2], i[3], i[4])
            _year = start_dt.year + 1 if eplusdt.standard_dt <= start_standard_dt else start_dt.year
            return eplusdt.datetime(_year)

        df.index = df.index.map(index_to_dt)
        df.sort(inplace=True)
        df = df.reindex(pd.date_range(df.index[0], df.index[-1], freq=self.freq))

        return df

    def set_df(self, value):
        if not isinstance(value, pd.DataFrame):
            raise EPWError("df must be a DataFrame")
        assert_index_equal(value.index, self._df.index)
        assert_index_equal(value.columns, self._df.columns)


def parse_epw(file_like, encoding=None):
    """
    Read and parse an EPW object as it was an EPW file

    Parameters
    ----------
    file_like
    encoding

    Returns
    -------
    first return is a dataframe and second is a header object (dataframe, header)
    """
    header_s = ""

    # header
    for line_s in file_like:
        header_s += line_s
        if re.match("^DATA PERIODS", line_s) is not None:
            break
    header = EPWHeader(header_s)

    # data
    df = pd.read_csv(file_like, header=None, low_memory=False, encoding=encoding)
    df.columns = EPW_COLUMNS[:len(df.columns)]
    df.index = pd.MultiIndex.from_arrays([df[c] for c in df.columns[:5]])
    for c in df.columns[:5].tolist():
        del df[c]

    return df, header




import platform
import re
import os
import logging


class ConfigurationError(Exception):
    pass


logger = logging.getLogger(__name__)


# ------------------------------------------------- DEFAULT VARIABLES --------------------------------------------------
class _Config:
    def __init__(self):
        self.eplus_available_versions = {}  # # {version: base_dir_path, ...
        self.eplus_version = None  # (i1, i2, i3)
        self.encoding = "latin-1"
        self.simulation_base_name = "oplus"
        self.default_write_style = "default write"
        self.default_read_style = None

    @property
    def eplus_base_dir_path(self):
        assert self.eplus_version is not None, "Eplus version must be set."
        assert self.eplus_version in self.eplus_available_versions, "Eplus version was not found in available versions."
        return self.eplus_available_versions[self.eplus_version]

    @property
    def os_name(self):
        global os_name
        return os_name


CONF = _Config()

# --------------------------------------------- END OF DEFAULT VARIABLES -----------------------------------------------

# set operating system
sys_name = platform.system()
if sys_name in ("Windows",):  # windows
    os_name = "windows"
elif sys_name in ("Darwin",):  # mac osx
    os_name = "osx"
elif sys_name in ("Linux", ):  # linux
    os_name = "linux"
else:
    raise ConfigurationError("Unknown platform.system(): '%s'." % sys_name)


# get systems specific configurations
if os_name == "windows":
    apps_dir, pattern = r"C:\\", re.compile("EnergyPlusV(\d*)-(\d*)-(\d*)")
elif os_name == "osx":  # mac osx
    apps_dir, pattern = "/Applications", re.compile("EnergyPlus-(\d*)-(\d*)-(\d*)")
elif os_name == "linux":  # linux
    apps_dir, pattern = "/usr/local", re.compile("EnergyPlus-(\d*)-(\d*)-(\d*)")
else:
    raise ConfigurationError("Unknown os_name: '%s'" % CONF.os_name)

# find most recent version of EnergyPlus
for file_name in os.listdir(apps_dir):
    match = pattern.search(file_name)
    if match is not None:
        version_tuple = (int(match.group(1)), int(match.group(2)), int(match.group(3)))
        CONF.eplus_available_versions[version_tuple] = os.path.join(apps_dir, file_name)

if len(CONF.eplus_available_versions) == 0:
    logger.warning("Base directory was not found. You must provide the path with 'set_configuration'.")
else:
    CONF.eplus_version = sorted(CONF.eplus_available_versions.keys(), reverse=True)[0]
    logger.info(
        "EnergyPlus version: '%s', base directory: '%s'. Use 'set_configuration' to change it." %
        (".".join([str(i) for i in CONF.eplus_version]), CONF.eplus_available_versions[CONF.eplus_version]))



"""
Simulation
------------


"""

import os
import shutil
from threading import Thread
import logging

from oplus.configuration import CONF
from oplus.util import run_subprocess, Enum, LoggerStreamWriter
from oplus.idf import IDF
from oplus.idd import IDD
from oplus.epw import EPW
from oplus.standard_output import StandardOutputFile
from oplus.mtd import MTD
from oplus.eio import EIO
from oplus.err import ERR
from oplus.summary_table import SummaryTable


class SimulationError(Exception):
    pass


class WrongExtensionError(SimulationError):
    pass


FILE_REFS = Enum(**dict((ref, ref) for ref in (
    "idf", "epw", "eio", "eso", "mtr", "mtd", "mdd", "err", "summary_table")))


class FileInfo:
    def __init__(self, constructor, get_path):
        self.constructor = constructor
        self.get_path = get_path


def get_input_file_path(dir_path, file_ref):
    assert file_ref in (FILE_REFS.idf, FILE_REFS.epw), "'%s' file ref is not an input file"
    return os.path.join(dir_path, "%s.%s" % (CONF.simulation_base_name, file_ref))


def get_common_output_file_path(dir_path, file_ref):
    assert file_ref in (FILE_REFS.eio, FILE_REFS.eso, FILE_REFS.mtr, FILE_REFS.mtd, FILE_REFS.mdd, FILE_REFS.err), \
        "'%s' file ref is not a common output file"

    if CONF.os_name == "windows":
        if CONF.eplus_version[:2] <= (8, 1):  # todo: check that it is not 8.2 or 8.3
            return os.path.join(dir_path, "%s.%s" % (CONF.simulation_base_name, file_ref))
        else:
            return os.path.join(dir_path, "eplusout.%s" % file_ref)
    elif CONF.os_name == "osx":
        if CONF.eplus_version[:2] <= (8, 1):  # todo: check that it is not 8.2 or 8.3
            return os.path.join(dir_path, "Output", "%s.%s" % (CONF.simulation_base_name, file_ref))
        else:
            if file_ref in ("idf", "epw"):
                return os.path.join(dir_path, "%s.%s" % (CONF.simulation_base_name, file_ref))
            else:
                return os.path.join(dir_path, "eplusout.%s" % file_ref)
    elif CONF.os_name == "linux":
        # todo : check all versions of Energyplus
        return os.path.join(dir_path, "Output", "%s.%s" % (CONF.simulation_base_name, file_ref))
    else:
        raise NotImplementedError("Linux not implemented yet.")


def get_summary_table_file_path(dir_path):
    # TODO: code and test properly
    if CONF.os_name == "windows":
        if CONF.eplus_version[:2] <= (8, 1):  # todo: check that it is not 8.2 or 8.3
            return os.path.join(dir_path, "%sTable.csv" % CONF.simulation_base_name)
        else:
            return os.path.join(dir_path, "eplustbl.csv")
    elif CONF.os_name == "osx":
        if CONF.eplus_version[:2] <= (8, 1):
            return os.path.join(dir_path, "Output", "%sTable.csv" % CONF.simulation_base_name)
        else:
            return os.path.join(dir_path, "eplustbl.csv")

    elif CONF.os_name == "linux":
        # todo : check all versions of Energyplus
        return os.path.join(dir_path, "Output", "%sTable.csv" % CONF.simulation_base_name)
    raise NotImplemented()


class Simulation:
    # for subclassing
    _idf_cls = IDF
    _idd_cls = IDD
    _epw_cls = EPW
    _standard_output_file_cls = StandardOutputFile
    _mtd_cls = MTD
    _eio_cls = EIO
    _summary_table_cls = SummaryTable
    _err_cls = ERR

    @classmethod
    def simulate(
            cls,
            idf_or_path,
            epw_or_path,
            base_dir_path,
            simulation_name=None,
            start=None,
            simulation_control=None,
            encoding=None,
            idd_or_path=None,
            stdout=None,
            stderr=None,
            beat_freq=None
    ):
        """
        simulation will be done in os.path.join(base_dir_path, simulation_name) if simulation has a name, else in
            base_dir_path
        default stdout and stderr are logger.info and logger.error
        """
        # manage simulation dir path
        assert os.path.isdir(base_dir_path), "Base dir path not found: '%s'" % base_dir_path
        simulation_dir_path = base_dir_path if simulation_name is None else os.path.join(base_dir_path, simulation_name)

        # make directory if does not exist
        if not os.path.exists(simulation_dir_path):
            os.mkdir(simulation_dir_path)

        # simulation control
        # todo: shouldn't this be in idf builder ?
        if simulation_control is not None:
            _sizing_ = "Sizing"
            _run_periods_ = "RunPeriods"
            if simulation_control not in (_sizing_, _run_periods_):
                raise SimulationError("Unknown simulation_control: '%s'. Must be in : %s." %
                                      (simulation_control, (_sizing_, _run_periods_)))
            if not isinstance(idf_or_path, IDF):
                idf_or_path = IDF(idf_or_path, encoding=encoding, idd_or_path=idd_or_path)

            sc = idf_or_path("SimulationControl").one
            if simulation_control == _sizing_:
                # prepare SimulationControl
                sc["Do Zone Sizing Calculation"] = "Yes"
                sc["Do System Sizing Calculation"] = "Yes"
                sc["Do Plant Sizing Calculation"] = "Yes"
                sc["Run Simulation for Sizing Periods"] = "Yes"
                sc["Run Simulation for Weather File Run Periods"] = "No"
            if simulation_control == _run_periods_:
                sc["Do Zone Sizing Calculation"] = "Yes"
                sc["Do System Sizing Calculation"] = "Yes"
                sc["Do Plant Sizing Calculation"] = "Yes"
                sc["Run Simulation for Sizing Periods"] = "No"
                sc["Run Simulation for Weather File Run Periods"] = "Yes"

        # run simulation
        stdout = LoggerStreamWriter(logger_name=__name__, level=logging.INFO) if stdout is None else stdout
        stderr = LoggerStreamWriter(logger_name=__name__, level=logging.ERROR) if stderr is None else stderr
        run_eplus(
            idf_or_path,
            epw_or_path,
            simulation_dir_path,
            stdout=stdout,
            stderr=stderr,
            beat_freq=beat_freq
        )

        # return simulation object
        return cls(
            base_dir_path,
            simulation_name=simulation_name,
            start=start,
            encoding=encoding,
            idd_or_path=idd_or_path
        )

    def __init__(
            self,
            base_dir_path,
            simulation_name=None,
            start=None,
            encoding=None,
            idd_or_path=None
    ):
        self._dir_path = base_dir_path if simulation_name is None else os.path.join(base_dir_path, simulation_name)
        self._start = start
        self._encoding = encoding
        self._idd_or_path = idd_or_path
        self.__idd = None
        self._file_refs = None

        # check simulation directory path exists
        assert os.path.isdir(self._dir_path), "Simulation directory does not exist: '%s'." % self._dir_path

    @property
    def file_refs(self):
        """
        Defined here so that we can use the class variables, in order to subclass in opluslus
        """
        if self._file_refs is None:
            self._file_refs = {
                FILE_REFS.idf: FileInfo(
                    constructor=lambda path: self._idf_cls(path, idd_or_path=self._idd, encoding=self._encoding),
                    get_path=lambda: get_input_file_path(self.dir_path, FILE_REFS.idf)
                ),

                FILE_REFS.epw: FileInfo(
                    constructor=lambda path: self._epw_cls(path, encoding=self._encoding, start=self._start),
                    get_path=lambda: get_input_file_path(self.dir_path, FILE_REFS.epw)
                ),

                FILE_REFS.eio: FileInfo(
                    constructor=lambda path: self._eio_cls(path, encoding=self._encoding),
                    get_path=lambda: get_common_output_file_path(self.dir_path, FILE_REFS.eio)
                ),

                FILE_REFS.eso: FileInfo(
                    constructor=lambda path: self._standard_output_file_cls(
                        path,
                        encoding=self._encoding,
                        start=self._start
                    ),
                    get_path=lambda: get_common_output_file_path(
                        self.dir_path,
                        FILE_REFS.eso
                    )
                ),

                FILE_REFS.mtr: FileInfo(
                    constructor=lambda path: self._standard_output_file_cls(
                        path,
                        encoding=self._encoding,
                        start=self._start),
                    get_path=lambda: get_common_output_file_path(self.dir_path, FILE_REFS.mtr)
                ),

                FILE_REFS.mtd: FileInfo(
                    constructor=lambda path: self._mtd_cls(path, encoding=self._encoding),
                    get_path=lambda: get_common_output_file_path(self.dir_path, FILE_REFS.mtd)
                ),

                FILE_REFS.mdd: FileInfo(
                    constructor=lambda path: open(path, encoding=self._encoding).read(),
                    get_path=lambda: get_common_output_file_path(self.dir_path, FILE_REFS.mdd)

                ),

                FILE_REFS.err: FileInfo(
                    constructor=lambda path: self._err_cls(path, encoding=self._encoding),
                    get_path=lambda: get_common_output_file_path(self.dir_path, FILE_REFS.err)
                ),

                FILE_REFS.summary_table: FileInfo(
                    constructor=lambda path: self._summary_table_cls(path, encoding=self._encoding),
                    get_path=lambda: get_summary_table_file_path(self.dir_path)
                )
            }
        return self._file_refs

    @property
    def dir_path(self):
        return self._dir_path

    @property
    def _idd(self):
        if self.__idd is None:
            self.__idd = IDD.get_idd(self._idd_or_path, encoding=self._encoding)
        return self.__idd

    def _check_file_ref(self, file_ref):
        assert file_ref in self._file_refs, "Unknown extension: '%s'." % file_ref

    def _path(self, file_ref):
        return self.file_refs[file_ref].get_path()

    def exists(self, file_ref):
        assert file_ref in FILE_REFS, \
            "Unknown file_ref: '%s'. Available: '%s'." % (file_ref, list(sorted(FILE_REFS.keys())))
        return os.path.isfile(self._path(file_ref))

    def path(self, file_ref):
        assert self.exists(file_ref), "File '%s' not found in simulation '%s'." % (file_ref, self._path(file_ref))
        return self._path(file_ref)

    def set_start(self, start):
        self._start = start

    def __getattr__(self, item):
        if item not in FILE_REFS:
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, item))

        return self.file_refs[item].constructor(self.path(item))


simulate = Simulation.simulate


def run_eplus(idf_or_path, epw_or_path, dir_path, stdout=None, stderr=None, beat_freq=None):
    """
    Parameters
    ----------
    idf_or_path
    epw_or_path
    dir_path
    stdout: default sys.stdout
    stderr: default sys.stderr
    beat_freq: if not none, stdout will be used at least every beat_freq (in seconds)
    """
    # work with absolute paths
    dir_path = os.path.abspath(dir_path)

    # check dir path
    if not os.path.isdir(dir_path):
        raise SimulationError("Simulation directory does not exist: '%s'." % dir_path)

    # save files
    simulation_idf_path = os.path.join(dir_path, CONF.simulation_base_name + ".idf")
    if isinstance(idf_or_path, IDF):
        idf_or_path.save_as(simulation_idf_path)
    else:
        shutil.copy2(idf_or_path, simulation_idf_path)

    simulation_epw_path = os.path.join(dir_path, CONF.simulation_base_name + ".epw")
    if isinstance(epw_or_path, EPW):
        epw_or_path.save_as(simulation_epw_path)
    else:
        shutil.copy2(epw_or_path, simulation_epw_path)

    # copy epw on windows (on linux or osx, epw may remain in current directory)
    if CONF.os_name == "windows":
        temp_epw_path = os.path.join(CONF.eplus_base_dir_path, "WeatherData", "%s.epw" % CONF.simulation_base_name)
        shutil.copy2(simulation_epw_path, temp_epw_path)
    else:
        temp_epw_path = None

    # prepare command
    if CONF.os_name == "windows":
        if CONF.eplus_version[:2] <= (8, 1):  # todo: check that it is not 8.2 or 8.3
            last_name = "RunEPlus.bat"
        else:
            last_name = "energyplus"
    elif CONF.os_name == "osx":
        if CONF.eplus_version[:2] <= (8, 1):  # todo: check that it is not 8.2 or 8.3
            last_name = "runenergyplus"
        else:
            last_name = "energyplus"
    elif CONF.os_name == "linux":
        if CONF.eplus_version[:2] < (8, 1):  # todo: check this limit is right
            last_name = "bin/runenergyplus"
        else:
            last_name = "runenergyplus"
    else:
        raise SimulationError("unknown os name: %s" % CONF.os_name)

    eplus_cmd = os.path.join(CONF.eplus_base_dir_path, last_name)

    # idf
    if CONF.os_name == "windows":
        if CONF.eplus_version[:2] <= (8, 1):  # todo: check that it is not 8.2 or 8.3
            idf_file_cmd = os.path.join(dir_path, CONF.simulation_base_name)
        else:
            idf_file_cmd = simulation_idf_path
    elif CONF.os_name == "osx":
        if CONF.eplus_version[:2] <= (8, 1):  # todo: check that it is not 8.2 or 8.3
            idf_file_cmd = os.path.join(dir_path, CONF.simulation_base_name)
        else:
            idf_file_cmd = simulation_idf_path
    elif CONF.os_name == "linux":
        idf_file_cmd = simulation_idf_path
    else:
        raise SimulationError("unknown os name: %s" % CONF.os_name)

    # epw
    if CONF.os_name == "windows":
        if CONF.eplus_version[:2] <= (8, 1):  # todo: check that it is not 8.2 or 8.3
            epw_file_cmd = CONF.simulation_base_name,  # only weather data name
        else:
            epw_file_cmd = simulation_epw_path
    elif CONF.os_name == "osx":
        epw_file_cmd = simulation_epw_path
    elif CONF.os_name == "linux":
        epw_file_cmd = simulation_epw_path
    else:
        raise SimulationError("unknown os name: %s" % CONF.os_name)

    # command list
    if CONF.os_name == "windows":
        if CONF.eplus_version[:2] <= (8, 1):  # todo: check that it is not 8.2 or 8.3
            cmd_l = [eplus_cmd, idf_file_cmd, epw_file_cmd]
        else:
            cmd_l = [eplus_cmd, "-w", epw_file_cmd, "-r", idf_file_cmd]
    elif CONF.os_name == "osx":
        if CONF.eplus_version[:2] <= (8, 1):  # todo: check that it is not 8.2 or 8.3
            cmd_l = [eplus_cmd, idf_file_cmd, epw_file_cmd]
        else:
            cmd_l = [eplus_cmd, "-w", epw_file_cmd, "-r", idf_file_cmd]
    elif CONF.os_name == "linux":
        cmd_l = [eplus_cmd, idf_file_cmd, epw_file_cmd]
    else:
        raise SimulationError("unknown os name: %s" % CONF.os_name)
    # launch calculation
    run_subprocess(
        cmd_l,
        cwd=dir_path,
        stdout=stdout,
        stderr=stderr,
        beat_freq=beat_freq
    )

    # if needed, we delete temp weather data (only on Windows, see above)
    if temp_epw_path is not None:
        os.remove(os.path.join(temp_epw_path))



"""
MTD
-------


"""

import os
import re

from oplus.configuration import CONF


class MTDError(Exception):
    pass


class MTD:
    def __init__(self, path, encoding=None):
        if not os.path.isfile(path):
            raise MTDError("No file at given path: '%s'." % path)
        self._path = path
        self._encoding = encoding

        self._variables_d, self._meters_d = self._parse()

    def _parse(self):
        variables_d = {}  # {ref: object, ...}
        meters_d = {}  # {ref: object, ...}

        output_l, current, current_s = None, None, None
        var_pattern = r"^ Meters for (\d*),([^\[\]]*) \[([\w\d]*)\]$"
        meter_pattern = r"^ For Meter=([^\[\]]*) \[([\w\d]*)],(.*)$"

        # build variables and meters
        with open(self._path, "r", encoding=CONF.encoding if self._encoding is None else self._encoding) as f:
            for line_s in f:
                if line_s == "\n":
                    if output_l is None:  # initialize
                        output_l = []
                    else:
                        output_l.append([current, current_s])
                    current, current_s = None, ""
                    continue
                if current is None:
                    # try variable
                    match = re.search(var_pattern, line_s)
                    if match is not None:
                        current = Variable(match.group(2), int(match.group(1)), match.group(3))
                        variables_d[current.ref] = current
                    else:
                        match = re.search(meter_pattern, line_s)
                        if match is None:
                            raise MTDError("Line was not parsed correctly: '%s'." % line_s)
                        kwargs = {}
                        for kv in line_s.split(",")[:-1]:
                            k, v = kv.split("=")
                            kwargs[k] = v
                        current = Meter(match.group(1), match.group(2), **kwargs)
                        meters_d[current.ref] = current
                else:
                    current_s += line_s

        # create links
        meter_pattern = r"^  OnMeter=([^\[\]]*) \[[\w\d]*\]$"
        var_pattern = r"^  (.*)$"
        for k, v in output_l:
            if isinstance(k, Variable):
                for v_s in v.split("\n")[:-1]:
                    match = re.search(meter_pattern, v_s)
                    if match is None:
                        raise MTDError("Meter pattern not parsed: '%s'." % v_s)
                    k.link_meter(meters_d[match.group(1)])
            else:
                for v_s in v.split("\n")[:-1]:
                    match = re.search(var_pattern, v_s)
                    if match is None:
                        raise MTDError("Variable pattern not parsed: '%s'." % v_s)
                    k.link_variable(variables_d[match.group(1)])

        return variables_d, meters_d

    def get_variable_refs(self, meter_ref):
        return [v.ref for v in self._meters_d[meter_ref].variables_l]

    def has_meter(self, meter_ref):
        return meter_ref in self._meters_d


class Meter:
    def __init__(self, ref, unit, **kwargs):
        self.ref = ref
        self.unit = unit
        self.kwargs = kwargs

        self.variables_l = []

    def link_variable(self, variable):
        if variable in self.variables_l:
            raise MTDError("Variable already linked.")
        self.variables_l.append(variable)


class Variable:
    def __init__(self, ref, variable_id, unit):
        self.ref = ref
        self.variable_id = variable_id
        self.unit = unit

        self.meters_l = []

    def link_meter(self, meter):
        if meter in self.meters_l:
            raise MTDError("Meter already linked.")
        self.meters_l = []

if __name__ == "__main__":
    mtd = MTD(r"C:\Users\Geoffroy\Desktop\simul_dir\rs_2013-opt_chicago.mtd")

    # print("VARIABLES")
    # for k in sorted(mtd._variables_d):
    #     print("'%s'" % k)
    #
    # print()
    # print("METERS")
    # for k in sorted(mtd._meters_d):
    #     print("'%s'" % k)

    # print(mtd.get_variable_refs("CH4:Facility"))
    meters_structure = [
        # "Heating", "Cooling", "InteriorLights", "ExteriorLights", "InteriorEquipment", "ExteriorEquipment", "Fans",
        # "Pumps", "HeatRejection", "Humidification", "HeatRecovery", "WaterSystems", "Refrigeration", "Generators"
        ["Electricity", ["Heating", "Cooling", "InteriorLights", "ExteriorLights", "InteriorEquipment", "Fans",
                         "HeatRecovery"]],
        ["Gas", ["Heating", "WaterSystems"]],
        ["Water", ["WaterSystems"]]
    ]
    for source, uses_l in meters_structure:
        for use in uses_l:
            meter_ref = "%s:%s" % (use, source)
            if not mtd.has_meter(meter_ref):
                print("ERROR: '%s' not found." % meter_ref)


import unittest
import os
import logging

from oplus.idf import IDF, IDFObject, IDFError
from oplus.util import CachingNotAllowedError
from oplus.idf import BrokenIDFError, IsPointedError
from oplus.configuration import CONF


logger = logging.getLogger(__name__)


schedule_test_object_str = """Schedule:Compact,
    %s,  !- Name
    Any Number,              !- Schedule Type Limits Name
    THROUGH: 12/31,          !- Field 1
    FOR: AllDays,            !- Field 2
    UNTIL: 12:00,1,          !- Field 3
    UNTIL: 24:00,0;          !- Field 5"""


class OneZoneEvapCooler(unittest.TestCase):
    """
    Tested under EPlus 8.1.0 on Windows (Geoffroy).
    !!! Only tests that do not modify IDF (avoid loading idf several times) - else use OneZoneEvapCoolerDynamic.
    """
    idf = None

    @classmethod
    def setUpClass(cls):
        cls.idf = IDF(os.path.join(CONF.eplus_base_dir_path, "ExampleFiles", "1ZoneEvapCooler.idf"))

    @classmethod
    def tearDownClass(cls):
        del cls.idf

    def test_idf_call(self):
        qs = self.idf("Construction")
        self.assertEqual({"R13WALL", "FLOOR", "ROOF31"}, set([c["name"] for c in qs]))

    def test_qs_one(self):
        self.assertEqual(self.idf("BuildingSurface:Detailed").filter("naMe", "Zn001:Roof001").one["name"],
                         "Zn001:Roof001")

    def test_idf_add_object(self):
        sch_name = "NEW TEST SCHEDULE"
        sch = self.idf.add_object(schedule_test_object_str % sch_name)
        self.assertTrue(isinstance(sch, IDFObject))

    def test_multi_level_filter(self):
        # get all building surfaces that have a zone with Z-Origin 0
        simple_filter_l = []
        for bsd in self.idf("BuildingSurface:Detailed"):
            if bsd["Zone name"][4] == 0:
                simple_filter_l.append(bsd)
        multi_filter_l = self.idf("BuildingSurface:Detailed").filter(("Zone Name", 4), 0).objects_l
        self.assertEqual(simple_filter_l, multi_filter_l)


class OneZoneEvapCoolerDynamic(unittest.TestCase):
    """
    Tested under EPlus 8.1.0 on Windows (Geoffroy).
    Here are tests that modify idf.
    """
    idf = None

    def setUp(self):
        self.idf = IDF(os.path.join(CONF.eplus_base_dir_path, "ExampleFiles", "1ZoneEvapCooler.idf"))

    @classmethod
    def tearDownClass(cls):
        del cls.idf

    def test_idf_add_object(self):
        sch_name = "NEW TEST SCHEDULE"
        self.idf.add_object(schedule_test_object_str % sch_name)
        self.assertEqual(self.idf("Schedule:Compact").filter("name", sch_name).one["name"], sch_name)

    def test_idf_remove_object(self):
        sch_name = "NEW TEST SCHEDULE"
        sch = self.idf.add_object(schedule_test_object_str % sch_name)
        self.idf.remove_object(sch)
        self.assertEqual(len(self.idf("Schedule:Compact").filter("name", sch_name)), 0)

    def test_idf_remove_object_raise(self):
        zone = self.idf("Zone").one
        self.assertRaises(IsPointedError, lambda: self.idf.remove_object(zone))

    def test_idf_remove_object_dont_raise(self):
        zone = self.idf("Zone").one
        self.idf.remove_object(zone, raise_if_pointed=False)
        self.assertEqual(len(self.idf("Zone")), 0)

    def test_pointing_objects(self):
        zone = self.idf("Zone").one
        self.assertEqual({"Zn001:Wall001", "Zn001:Wall002", "Zn001:Wall003", "Zn001:Wall004", "Zn001:Flr001",
                          "Zn001:Roof001"},
                         set([bsd["name"] for bsd in zone.pointing_objects("BuildingSurface:Detailed")]))

    def test_pointed_objects(self):
        bsd = self.idf("BuildingSurface:Detailed").filter("name", "Zn001:Wall001").one
        zone = self.idf("Zone").filter("name", "Main Zone").one
        self.assertEqual(bsd["zone name"], zone)
        self.assertEqual(bsd[3], zone)

    def test_idf_copy(self):
        old_name = "System Availability Schedule"
        old = self.idf("Schedule:Compact").filter("name", old_name).one
        new = old.copy()
        new_name = old_name + "- new"
        new["name"] = new_name
        self.assertNotEqual(old, new)

    def test_set_object_simple(self):
        new_name = "Fan Availability Schedule - 2"
        supply_fan = self.idf("Fan:ConstantVolume").filter("name", "Supply Fan").one
        supply_fan["availability schedule name"] = schedule_test_object_str % new_name
        # check set
        self.assertEqual(
            self.idf("Fan:ConstantVolume").filter("name", "Supply Fan").one["AvaiLABIlity schedule name"]["name"],
            new_name)

    def test_set_object_broken(self):
        supply_fan = self.idf("Fan:ConstantVolume").filter("name", "Supply Fan").one
        name = supply_fan["availability schedule name"]["name"]

        def raise_if_you_care():
            supply_fan["availability schedule name"] = schedule_test_object_str % name
        self.assertRaises(BrokenIDFError, raise_if_you_care)

    def test_extensible(self):
        sch = self.idf("Schedule:Compact").filter("name", "System Availability Schedule").one
        for i in range(1500):
            sch.add_field("12:00")
        self.assertEqual(sch[1300], "12:00")

    @unittest.skip("skipped but must be fixed !!!")
    def test_big_extensible(self):
        """
        Doesn't work but bypassed !!! todo: repair

        """
        sch = self.idf("Schedule:Compact").filter("name", "System Availability Schedule").one
        for i in range(4560):  # 4500 is the limit of idd
            sch.add_field("12:00")
        self.assertEqual(4560, len(sch))

    def test_pop_end(self):
        sch = self.idf("Schedule:Compact").filter("name", "System Availability Schedule").one
        ini_len = len(sch)
        self.assertEqual("1", sch.pop())
        self.assertEqual(ini_len-1, len(sch))

    def test_pop_middle(self):
        sch = self.idf("Schedule:Compact").filter("name", "System Availability Schedule").one

        self.assertEqual(sch.to_str(), """Schedule:Compact,
    System Availability Schedule,  ! - Name
    Any Number,                    ! - Schedule Type Limits Name
    THROUGH: 12/31,                ! - Field 1
    FOR: AllDays,                  ! - Field 2
    UNTIL: 24:00,                  ! - Field 3
    1;                             ! - Field 3
""")

        self.assertEqual("THROUGH: 12/31", sch.pop(2))
        self.assertEqual(sch.to_str(), """Schedule:Compact,
    System Availability Schedule,  ! - Name
    Any Number,                    ! - Schedule Type Limits Name
    FOR: AllDays,                  ! - Field 2
    UNTIL: 24:00,                  ! - Field 3
    1;                             ! - Field 3
""")

    def test_pop_raises(self):
        sch = self.idf("Schedule:Compact").filter("name", "System Availability Schedule").one
        self.assertRaises(IDFError, lambda: sch.pop(1))

    def test_cache_with_modification_raises(self):
        self.idf.activate_cache()
        sch = self.idf("Schedule:Compact").filter("name", "System Availability Schedule").one
        self.assertRaises(CachingNotAllowedError, lambda: sch.pop(1))

    def test_cache_on_filter(self):
        # activate
        self.idf.activate_cache()
        sch = self.idf("Schedule:Compact").filter("name", "System Availability Schedule").one
        self.assertEqual(1, len(self.idf._.cache))
        k = list(self.idf._.cache.keys())[0]
        self.assertEqual(0, self.idf._.cache[k]["hits"])

        # clear
        self.idf.clear_cache()
        self.assertEqual(0, len(self.idf._.cache))

        # retry
        sch = self.idf("Schedule:Compact").filter("name", "System Availability Schedule").one
        self.assertEqual(1, len(self.idf._.cache))
        k = list(self.idf._.cache.keys())[0]
        self.assertEqual(0, self.idf._.cache[k]["hits"])

        # check 1 hit
        sch = self.idf("Schedule:Compact").filter("name", "System Availability Schedule").one
        self.assertEqual(1, self.idf._.cache[k]["hits"])

        # deactivate
        self.idf.deactivate_cache()
        self.assertIs(None, self.idf._.cache)


class FourZoneWithShadingSimple1(unittest.TestCase):
    """
    Tested under EPlus 8.1.0 on Windows (Geoffroy).
    """
    idf = None

    @classmethod
    def tearDownClass(cls):
        del cls.idf

    def test_read_idf(self):
        self.idf = IDF(os.path.join(CONF.eplus_base_dir_path, "ExampleFiles", "4ZoneWithShading_Simple_1.idf"))


class FiveZoneAirCooled(unittest.TestCase):
    """
    Tested under EPlus 8.1.0 on Windows (Geoffroy).
    Tested under EPlus 8.1.0 on Mac (Antoine).
    """
    def test_multiple_branch_links(self):
        idf = IDF(os.path.join(CONF.eplus_base_dir_path, "ExampleFiles", "5ZoneAirCooled.idf"))
        bl = idf("BranchList").filter("Name", "Heating Supply Side Branches").one
        b3 = idf("Branch").filter("Name", "Heating Supply Bypass Branch").one
        self.assertEqual(bl[3], b3)



from setuptools import setup, find_packages
from pkg_resources import parse_requirements
import os


with open(os.path.join("oplus", "version.py")) as f:
    version = f.read().split("=")[1].strip().strip("'").strip('"')


def _get_req_list(file_name):
    with open(file_name) as f:
        return [str(r) for r in parse_requirements(f.read())]


setup(
    name='oplus',

    version=version,

    packages=find_packages(),

    author="Geoffroy d'Estaintot",

    author_email="geoffroy.destaintot@openergy.fr",

    description="A python package for working with Energy Plus",

    long_description=open('README.md').read(),  # long_description,

    install_requires=_get_req_list("requirements-conda.txt") + _get_req_list("requirements-pip.txt"),

    url='https://github.com/Openergy/oplus',

    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: French",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.4",
    ],

    keywords=['data', 'simulation'],

    package_data={'oplus': ['*.txt']},

    include_package_data=True
)



"""
Output table
---------------


"""
import os

import pandas as pd

from oplus.configuration import CONF


class OutputTableError(Exception):
    pass


class OutputTable:
    def __init__(self, path, encoding=None):
        if not os.path.isfile(path):
            raise OutputTableError("No file at given path: '%s'." % path)
        self._path = path
        self._encoding = encoding

        self._reports_d = self._parse()  # {report_name: {table_name: df, ...}, ...}

    def _parse(self):
        # constants
        _name_ = "name"
        _columns_ = "columns"
        _values_ = "values"
        _index_ = "index"

        # variables
        raw_reports_d = {}  # {report_name: [tables_d, ...], ...}
        current_raw_tables_l = None
        current_raw_table_d = None  # {"name": "name", "columns": "columns", values: [[v1, ...], ...]}
        columns_nb = None

        def to_float_if_possible(s):
            try:
                return float(s)
            except ValueError:
                if s.strip() == "":
                    return None
                else:
                    return s

        # loop
        with open(self._path, "r", encoding=CONF.encoding if self._encoding is None else self._encoding) as f:
            while True:
                # next line
                try:
                    line_s = next(f).strip()
                except StopIteration:
                    break

                # use everything except table names and values
                if line_s[:6] == "REPORT":
                    # find report name
                    report_name = line_s.split(",")[1].strip()
                    # create new report
                    current_raw_tables_l = []
                    raw_reports_d[report_name] = current_raw_tables_l
                    # create empty current_raw_table_d to initialize
                    current_raw_table_d = {_index_: [], _values_: []}
                    # skip two next lines
                    for i in range(2):
                        next(f)
                    continue
                elif current_raw_tables_l is None:
                    # first table not reached yet, nothing to do
                    continue
                elif line_s[:5] == "Note ":  # end notes
                    break

                # parse tables
                if line_s.strip() == "":
                    if _columns_ in current_raw_table_d:
                        # end of data, we create a new current_raw_table_d
                        current_raw_table_d = {_index_: [], _values_: []}
                elif _name_ not in current_raw_table_d:
                    # we know this table exists (we are not in end of file), so we name and append
                    current_raw_table_d[_name_] = line_s
                    current_raw_tables_l.append(current_raw_table_d)
                elif _columns_ not in current_raw_table_d:
                    columns_l = line_s.split(",")[2:]
                    current_raw_table_d[_columns_] = columns_l
                    columns_nb = len(columns_l)
                else:
                    line_l = line_s.split(",")
                    if len(line_l) <= 1:  # comments sometimes follow a table, without a whitespace
                        continue
                    current_raw_table_d[_index_].append(",".join(line_l[1:-columns_nb]))
                    current_raw_table_d[_values_].append([to_float_if_possible(s) for s in line_l[-columns_nb:]])

        # create dataframes
        reports_d = {}
        for report_name, raw_tables_l in raw_reports_d.items():
            tables_d = {}
            for raw_table_d in raw_tables_l:
                tables_d[raw_table_d[_name_]] = pd.DataFrame(data=raw_table_d[_values_], index=raw_table_d[_index_],
                                                             columns=raw_table_d[_columns_])
            reports_d[report_name] = tables_d

        return reports_d

    def get_table(self, table_name, report_name=None):
        if report_name is None:
            for rp_name, tables_d in self._reports_d.items():
                if table_name in tables_d:
                    return tables_d[table_name]
            raise OutputTableError("Table name '%s' not found." % table_name)

        if not report_name in self._reports_d:
            raise OutputTableError("Report name '%s' not found." % report_name)
        tables_d = self._reports_d[report_name]
        if not table_name in tables_d:
            raise OutputTableError("Table name '%s' not found in report '%s'." % (table_name, report_name))
        return tables_d[table_name]


if __name__ == "__main__":
    table = OutputTable(r"C:\Users\Geoffroy\Desktop\simul_dir\oplusTable.csv")
    #print(table._reports_d["Annual Building Utility Performance Summary"])
    print(table.get_table("End Uses"))





"""
IDD
----
Manages the EPlus idd file.

Definitions
-----------
object: EPlus object
field: EPlus object field
object descriptor: description of an object, given by the idd. An object descriptor has a unique ref (given by idd)
field descriptor: object descriptor field

pointing object (has tag 'object-list'): object that points towards another object
pointed object (has tag 'reference'): object being pointed by another object

"""

import os
import re
import logging

from oplus.configuration import CONF


logger = logging.getLogger(__name__)


class IDDError(Exception):
    pass


def get_idd_path():
    return os.path.join(CONF.eplus_base_dir_path, "Energy+.idd")


class FieldDescriptor:
    """
    No checks implemented (idf is considered as ok).
    """
    def __init__(self, field_basic_type, name=None):
        if field_basic_type not in ("A", "N"):
            raise IDDError("Unknown field type: '%s'." % field_basic_type)
        self.basic_type = field_basic_type  # A -> alphanumeric, N -> numeric
        self.name = name
        self._tags_d = {}

        self._detailed_type = None

    @property
    def tags(self):
        return sorted(self._tags_d)

    def add_tag(self, ref, value=None):
        if ref not in self._tags_d:
            self._tags_d[ref] = []
        if value is not None:
            self._tags_d[ref].append(value)

    def get_tag(self, ref):
        """
        Returns tag belonging to field descriptor. If 'note', will be string, else list of elements.
        """
        if ref == "note":  # memo is for object descriptors
            return " ".join(self._tags_d[ref])
        return self._tags_d[ref]

    def has_tag(self, ref):
        return ref in self._tags_d

    def basic_parse(self, value):
        """
        Parses raw value (string or None) to string, int, float or None.
        """
        # no value
        if (value is None) or (value == ""):
            return None

        # alphabetical
        if self.basic_type == "A":
            return str(value)

        # numeric
        if type(value) is str:
            if value.lower() in ("autocalculate", "autosize"):
                return value
            try:
                return int(value)
            except ValueError:
                return float(value)

        # in case has already been parsed
        elif type(value) in (int, float):
            return value

    @property
    def detailed_type(self):
        """
        Uses EPlus double approach of type ('type' tag, and/or 'key', 'object-list', 'external-list', 'reference' tags)
        to determine detailed type.
        Returns
        -------
        "integer", "real", "alpha", "choice", "reference", "object-list", "external_list", "node"
        """
        if self._detailed_type is None:
            if "reference" in self._tags_d:
                self._detailed_type = "reference"
            elif "type" in self._tags_d:
                self._detailed_type = self._tags_d["type"][0].lower()
            elif "key" in self._tags_d:
                self._detailed_type = "choice"
            elif "object-list" in self._tags_d:
                self._detailed_type = "object-list"
            elif "external-list" in self._tags_d:
                self._detailed_type = "external-list"
            elif self.basic_type == "A":
                self._detailed_type = "alpha"
            elif self.basic_type == "N":
                self._detailed_type = "real"
            else:
                raise IDDError("Can't find detailed type.")
        return self._detailed_type

    @staticmethod
    def name_to_formatted_name(name):
        """
        Returns formatted name of variable (lower case).
        """
        return name.lower()


class ObjectDescriptor:
    """
    Describes a EPlus object (see idd).
    """
    def __init__(self, ref, group_name=None):
        self.ref = ref
        self.group_name = group_name
        self._fieldds_l = []
        self._tags_d = {}

        # extensible management
        self._extensible_cycle_len = 0  # if 0: not extensible
        self._extensible_cycle_start = None  # will be filled first time asked

    @property
    def tags(self):
        return sorted(self._tags_d)

    @property
    def field_descriptors_l(self):
        return self._fieldds_l

    def get_tag(self, ref):
        """
        Returns tag belonging to object descriptor. If 'memo', will be string, else list of elements.
        """
        if ref == "memo":  # note if for field descriptors
            return " ".join(self._tags_d[ref])
        return self._tags_d[ref]

    def add_tag(self, ref, value=None):
        if value is None:
            return None

        if not ref in self._tags_d:
            self._tags_d[ref] = []
        self._tags_d[ref].append(value)

        # manage extensible
        if "extensible" in ref:
            self._extensible_cycle_len = int(ref.split(":")[1])

    def add_field_descriptor(self, field):
        """
        Adds a new field descriptor.
        """
        self._fieldds_l.append(field)

    def get_field_descriptor(self, index_or_name):
        """
        Returns
        -------
        asked field descriptor.
        """
        index = self.get_field_index(index_or_name)

        if index >= len(self._fieldds_l) and (self._extensible_cycle_len != 0):  # extensible object, find modulo
            index = self._extensible_cycle_len + ((index - self._extensible_cycle_start) % self._extensible_cycle_len)

        return self._fieldds_l[index]

    def get_field_index(self, index_or_name):
        """
        if index, must be >=0
        """
        # if index
        if type(index_or_name) is int:
            if index_or_name >= len(self._fieldds_l) and (self._extensible_cycle_len == 0):
                raise IDDError("Index out of range : %i." % index_or_name)
            return index_or_name
        # if name (extensible can not be used here)
        formatted_name = FieldDescriptor.name_to_formatted_name(index_or_name)
        for i, cur_field in enumerate(self._fieldds_l):
            cur_formatted_name = FieldDescriptor.name_to_formatted_name(cur_field.name)
            if cur_formatted_name == formatted_name:
                return i
        raise IDDError("No field of '%s' is named '%s'." % (self.ref, index_or_name))

    @property
    def formatted_ref(self):
        return self.ref.replace(":", "_")

    @property
    def extensible(self):
        """
        Returns cycle_len, cycle_start
        """
        if self._extensible_cycle_len == 0:
            return None, None
        if self._extensible_cycle_start is None:
            for i, fieldd in enumerate(self._fieldds_l):
                if fieldd.has_tag("begin-extensible"):
                    break
            else:
                raise IDDError("begin-extensible tag not found.")
            self._extensible_cycle_start = i
        return self._extensible_cycle_len, self._extensible_cycle_start

class IDD:
    @classmethod
    def get_idd(cls, idd_or_path, encoding=None):
        if idd_or_path is None:
            return cls()
        if isinstance(idd_or_path, str):
            return cls(path=idd_or_path, encoding=encoding)
        elif isinstance(idd_or_path, cls):
            return idd_or_path
        raise IDDError("'idd_or_path' must be a path or an IDD. Given object: '%s', type: '%s'." %
                       (idd_or_path, type(idd_or_path)))

    def __init__(self, path=None, encoding=None):
        if (path is not None) and not os.path.exists(path):
            raise IDDError("No file at given path: '%s'." % path)
        self.path = get_idd_path() if path is None else path
        self._encoding = encoding
        # od: object descriptor, linkd: link descriptor
        self._ods_d = {}  # object descriptors {lower_object_descriptor_ref: od, ...}
        self._pointed_od_linkds_d = {}  # linkd: link descriptor {link_name: [(od, field_index), ...], ...}
        self._pointing_od_linkds_d = {}  # {link_name: [(od, field_index), ...], ...}
        self._groups_d = {}  # {group name: [od, od, ...]}

        self._parse()
        self._link()

    def pointed_links(self, link_name):
        """
        Returns all the possible links named 'link_name' to pointed objects. A link is a combination of an object
        descriptor and an index. This corresponds to fields having a 'reference' tag.

        Returns
        -------
        list of links: [(object_descriptor_ref, index), ...]
        """
        if not link_name in self._pointed_od_linkds_d:
            logger.info("Idd useless ref -> '%s' ref is defined, but no object-list pointing (idd problem, "
                        "nothing to do)." % link_name)
            return []
        return self._pointed_od_linkds_d[link_name]

    def pointing_links(self, link_name):
        """
        Returns all the possible links named 'link_name' to pointing objects. A link is a combination of an object
        descriptor and an index. This corresponds to a field having an 'object-list' tag.

        Returns
        -------
        list of links: [(object_descriptor_ref, index), ...]
        """
        if not link_name in self._pointing_od_linkds_d:
            logger.debug("No pointing links ('object-list') with name '%s'. This may be an idd bug, or a wrong "
                         "link_name may have been provided." % link_name)
            return []
        return self._pointing_od_linkds_d[link_name]

    @property
    def groups_l(self):
        """
        All group names.
        """
        return sorted(self._groups_d)

    def _parse(self):
        """ Parses idd file."""
        group_name, od, fieldd = None, None, None
        with open(self.path, "r", encoding=CONF.encoding if self._encoding is None else self._encoding) as f:
            for i, raw_line in enumerate(f):
                line = raw_line.split("!")[0]  # we tear comment

                # blank line
                if re.search(r"^\s*$", line) is not None:
                    continue

                # group comment (must be before comments)
                match = re.search(r"^\\group (.+)$", line)
                if match is not None:
                    group_name = match.group(1).strip()
                    self._groups_d[group_name] = []
                    # re-initialize
                    od, fieldd = None, None
                    continue

                # tag
                match = re.search(r"^\s*\\(.+)$", line)
                if match is not None:
                    # identify
                    content = match.group(1)
                    if not " " in content:  # only a ref
                        tag_ref = content.strip()
                        tag_value = None
                    else:  # ref and value
                        match = re.search(r"^([\w\-\>\<:]+) (.*)$", content)
                        tag_ref = match.group(1)
                        tag_value = match.group(2).strip()
                    # store
                    if fieldd is None:  # we are not in a field -> object descriptor comment
                        od.add_tag(tag_ref, tag_value)
                    else:  # we are in a field
                        fieldd.add_tag(tag_ref, tag_value)
                    continue

                # named field descriptor
                match = re.search(r"^\s*([AN])\d+\s*([;,])\s*\\[fF]ield (.*)$", line)
                if match is not None:
                    # identify
                    fieldd_type = match.group(1)
                    name = match.group(3).strip()
                    if name == "":
                        name = None
                    # store
                    fieldd = FieldDescriptor(fieldd_type, name=name)
                    od.add_field_descriptor(fieldd)
                    continue

                # unnamed field descriptors
                match = re.search(r"^\s*[AN]\d+.+([;,])\s*\\note.*$", line)
                if match is not None:
                    # identify
                    fields_l = [s.strip() for s in line.split(r"\note")[0].strip()[:-1].split(",")]
                    for fieldd_s in fields_l:
                        fieldd_type = fieldd_s[0]
                        # store
                        fieldd = FieldDescriptor(fieldd_type)
                        od.add_field_descriptor(fieldd)
                    continue

                # od: object descriptor
                match = re.search(r"^\s*([\w:\-]+),\s*$", line)
                if match is not None:
                    # identify
                    ref = match.group(1).strip()
                    if group_name is None:
                        raise IDDError("No group name.")
                    # store
                    od = ObjectDescriptor(ref, group_name=group_name)
                    if ref in self._ods_d:
                        raise IDDError("Object descriptor already registered.")
                    self._ods_d[ref.lower()] = od
                    self._groups_d[group_name].append(od)

                    # re-initialize
                    fieldd = None
                    continue

                # non parsed line - special objects
                if ("Lead Input;" in line) or ("Simulation Data;" in line):
                    # store
                    ref = line.strip()[:-1]
                    self._ods_d[ref.lower()] = ObjectDescriptor(ref)  # start
                    self._ods_d["end " + ref.lower()] = ObjectDescriptor("End " + ref)  # end
                    # re-initialize
                    od, fieldd = None, None
                    continue

                raise IDDError("Line %i not parsed: '%s'." % (i+1, raw_line))

    def _link(self):
        """ Links object descriptors together. """
        for od in self._ods_d.values():
            for i, fieldd in enumerate(od.field_descriptors_l):
                if fieldd.has_tag("reference"):
                    for ref_name in fieldd.get_tag("reference"):
                        if not ref_name in self._pointed_od_linkds_d:
                            self._pointed_od_linkds_d[ref_name] = []
                        self._pointed_od_linkds_d[ref_name].append((od, i))
                if fieldd.has_tag("object-list"):
                    for ref_name in fieldd.get_tag("object-list"):
                        if not ref_name in self._pointing_od_linkds_d:
                            self._pointing_od_linkds_d[ref_name] = []
                        self._pointing_od_linkds_d[ref_name].append((od, i))

    def get_object_descriptor(self, od_ref):
        """
        Arguments
        ---------
        od_ref: object descriptor reference.

        Returns
        -------
        object descriptor
        """

        return self._ods_d[od_ref.lower()]

    def get_object_descriptors_by_group(self, group_name):
        """
        Returns
        -------
        list of object descriptors belonging to a given group.
        """
        return self._groups_d[group_name]